/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Provides access to the Network Interface Unit h/w block.
 */
#include <unm_inc.h>
#include <linux/kernel.h>
#include <asm/string.h> /* for memset */
#include <linux/delay.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
#include <linux/hardirq.h>
#else 
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#include <asm/hardirq.h>
#endif
#endif
#include <asm/processor.h>

#include "unm_nic.h"

#ifndef in_atomic
#define in_atomic()     (1)
#endif

static long phy_lock_timeout= 100000000;

static int phy_lock(struct unm_adapter_s *adapter)
{
    int i;
    int done = 0, timeout = 0;

    while (!done) {
        /* acquire semaphore3 from PCI HW block */
        unm_nic_hw_read_wx(adapter, UNM_PCIE_REG(PCIE_SEM3_LOCK), &done, 4);
        if (done == 1)
            break;
        if (timeout >= phy_lock_timeout) {
            return -1;
        }
        timeout++;
        /*
         * Yield CPU
         */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        if(!in_atomic())
                schedule();
        else {
#endif
                for(i = 0; i < 20; i++)
                        cpu_relax();    /*This a nop instr on i386*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        }
#endif
    }
    unm_crb_writelit_adapter(adapter, UNM_PHY_LOCK_ID, PHY_LOCK_DRIVER);
    return 0;
}

static void phy_unlock(struct unm_adapter_s *adapter)
{
    int val;
    /* release semaphore3 */
    unm_nic_hw_read_wx(adapter, UNM_PCIE_REG(PCIE_SEM3_UNLOCK), &val, 4);
}

/* 
 * unm_niu_gbe_phy_read - read a register from the GbE PHY via
 * mii management interface.
 *
 * Note: The MII management interface goes through port 0.
 *       Individual phys are addressed as follows:
 *       [15:8]  phy id
 *       [7:0]   register number
 *
 * Returns:  0 success
 *          -1 error
 *
 */
long unm_niu_gbe_phy_read (struct unm_adapter_s *adapter,
			   long phy, long reg, unm_crbword_t *readval) {
    unm_niu_gb_mii_mgmt_address_t address;
    unm_niu_gb_mii_mgmt_command_t command;
    unm_niu_gb_mii_mgmt_indicators_t status;

    long timeout=0;
    long result = 0;
    long restore = 0;
    unm_niu_gb_mac_config_0_t mac_cfg0;

    if (phy_lock(adapter) != 0) {
        return -1;
    }

    /* MII mgmt all goes through port 0 MAC interface, so it cannot be in reset */
    unm_nic_hw_read_wx(adapter, UNM_NIU_GB_MAC_CONFIG_0(0), &mac_cfg0, 4);
    if (mac_cfg0.soft_reset) {
        unm_niu_gb_mac_config_0_t temp;
        *(unm_crbword_t *)&temp = 0;    
        temp.tx_reset_pb = 1;
        temp.rx_reset_pb = 1;
        temp.tx_reset_mac = 1;
        temp.rx_reset_mac = 1;
        unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MAC_CONFIG_0(0), &temp, 4);
        restore = 1;
    }


    *(unm_crbword_t *)&address = 0;
    address.reg_addr = reg;
    address.phy_addr = phy;
    unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MII_MGMT_ADDR(0), &address, 4);
    
    *(unm_crbword_t *)&command = 0;    /* turn off any prior activity */
    unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MII_MGMT_COMMAND(0), &command, 4);

    /* send read command */
    command.read_cycle=1;
    unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MII_MGMT_COMMAND(0), &command, 4);

    *(unm_crbword_t *)&status = 0;   
    do {
        unm_nic_hw_read_wx(adapter, UNM_NIU_GB_MII_MGMT_INDICATE(0),&status, 4);
        timeout++;
    } while ((status.busy || status.notvalid) && (timeout++ < UNM_NIU_PHY_WAITMAX));

    if (timeout < UNM_NIU_PHY_WAITMAX) {
        unm_nic_hw_read_wx(adapter, UNM_NIU_GB_MII_MGMT_STATUS(0),readval, 4);
        result = 0;
    }else {
        result = -1;
    }

    if (restore) 
        unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MAC_CONFIG_0(0), &mac_cfg0, 4);

    phy_unlock(adapter);

    return (result);

}
/* 
 * unm_niu_gbe_phy_write - write a register to the GbE PHY via
 * mii management interface.
 *
 * Note: The MII management interface goes through port 0.
 *       Individual phys are addressed as follows:
 *       [15:8]  phy id
 *       [7:0]   register number
 *
 * Returns:  0 success
 *          -1 error
 *
 */
long unm_niu_gbe_phy_write (struct unm_adapter_s *adapter,
			    long phy, long reg, unm_crbword_t val) {
    unm_niu_gb_mii_mgmt_address_t address;
    unm_niu_gb_mii_mgmt_command_t command;
    unm_niu_gb_mii_mgmt_indicators_t status;
    long timeout=0;
    long result = 0;
    long restore = 0;
    unm_niu_gb_mac_config_0_t mac_cfg0;

    if (phy_lock(adapter) != 0) {
        return -1;
    }

    /* MII mgmt all goes through port 0 MAC interface, so it cannot be in reset */
    unm_nic_hw_read_wx(adapter, UNM_NIU_GB_MAC_CONFIG_0(0), &mac_cfg0, 4);
    if (mac_cfg0.soft_reset) {
        unm_niu_gb_mac_config_0_t temp;
        *(unm_crbword_t *)&temp = 0;    
        temp.tx_reset_pb = 1;
        temp.rx_reset_pb = 1;
        temp.tx_reset_mac = 1;
        temp.rx_reset_mac = 1;
        unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MAC_CONFIG_0(0), &temp , 4);
        restore = 1;
    }

    *(unm_crbword_t *)&command = 0;    /* turn off any prior activity */
    unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MII_MGMT_COMMAND(0), &command, 4);

    *(unm_crbword_t *)&address = 0;
    address.reg_addr = reg;
    address.phy_addr = phy;
    unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MII_MGMT_ADDR(0), &address, 4);
    
    unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MII_MGMT_CTRL(0), &val, 4);


    *(unm_crbword_t *)&status = 0;   
    do {
        unm_nic_hw_read_wx(adapter, UNM_NIU_GB_MII_MGMT_INDICATE(0),&status, 4);
        timeout++;
    } while ((status.busy) && (timeout++ < UNM_NIU_PHY_WAITMAX));

    if (timeout < UNM_NIU_PHY_WAITMAX) {
        result = 0;
    }else {
        result = -1;
    }

    /* restore the state of port 0 MAC in case we tampered with it */
    if (restore) 
        unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MAC_CONFIG_0(0), &mac_cfg0, 4);

    phy_unlock(adapter);

    return (result);

}

long unm_niu_gbe_enable_phy_interrupts(struct unm_adapter_s *adapter,
				       long port) {
    long result = 0;
    unm_niu_phy_interrupt_t enable;
    *(unm_crbword_t *)&enable = 0;
    enable.link_status_changed = 1;   
    enable.autoneg_completed= 1;   
    enable.speed_changed = 1;
    if (0 != unm_niu_gbe_phy_write(adapter, port,
				   UNM_NIU_GB_MII_MGMT_ADDR_INT_ENABLE,
				   *(unm_crbword_t *)&enable)) {
        result = -1;
    }
    return result;
}

long unm_niu_gbe_disable_phy_interrupts(struct unm_adapter_s *adapter,
					long port) {
    long result = 0;
    if (0 != unm_niu_gbe_phy_write(adapter, port,
				   UNM_NIU_GB_MII_MGMT_ADDR_INT_ENABLE,0)) {
        result = -1;
    }
    return result;
}
long unm_niu_gbe_clear_phy_interrupts(struct unm_adapter_s *adapter,
				      long port) {
    long result = 0;
    if (0 != unm_niu_gbe_phy_write(adapter, port,
				   UNM_NIU_GB_MII_MGMT_ADDR_INT_STATUS,-1)) {
        result = -1;
    }
    return result;
}


/*
 * Return the current station MAC address.
 * Note that the passed-in value must already be in network byte order.
 */
int
unm_niu_macaddr_get(struct unm_adapter_s *adapter,
                    int phy, unsigned char *addr)
{
    __uint64_t result;
    unsigned long flags;

    if (addr == NULL)
        return -1;
    if ((phy < 0) || (phy > 3))
        return -1;

    write_lock_irqsave(&adapter->adapter_lock, flags);
    if (adapter->curr_window != 0) {
	unm_nic_pci_change_crbwindow(adapter, 0);
    }
    result = readl((void *)pci_base_offset(adapter, UNM_NIU_GB_STATION_ADDR_1(phy)))
                                                                  >> 16;
    result |= ((uint64_t)readl((void *)pci_base_offset(adapter,
                                       UNM_NIU_GB_STATION_ADDR_0(phy)))) << 16;

    memcpy(addr, &result, sizeof(unm_ethernet_macaddr_t));

    unm_nic_pci_change_crbwindow(adapter, 1);

    write_unlock_irqrestore(&adapter->adapter_lock, flags);

    return 0;
}

/*
 * Set the station MAC address.
 * Note that the passed-in value must already be in network byte order.
 */
int
unm_niu_macaddr_set(struct unm_adapter_s *adapter,
                    int phy, unm_ethernet_macaddr_t addr)
{
    unm_crbword_t temp = 0;
    unsigned long flags;

    if ((phy < 0) || (phy > 3))
        return -1;

    memcpy(&temp,addr,2);
    temp <<= 16;
    write_lock_irqsave(&adapter->adapter_lock, flags);
    if (adapter->curr_window != 0) {
	unm_nic_pci_change_crbwindow(adapter, 0);
    }
    writel(*(__uint32_t *)&temp, (void *)pci_base_offset(adapter, UNM_NIU_GB_STATION_ADDR_1(phy)));

    temp = 0;

    memcpy(&temp,((__uint8_t *)addr)+2,sizeof(unm_crbword_t));
    writel(*(__uint32_t *)&temp, (void *)pci_base_offset(adapter, UNM_NIU_GB_STATION_ADDR_0(phy)));
    unm_nic_pci_change_crbwindow(adapter, 1);
    write_unlock_irqrestore(&adapter->adapter_lock, flags);

    return 0;
}

/* Disable a GbE interface */
native_t unm_niu_disable_gbe_port(struct unm_adapter_s *adapter,
				  native_t port) {
    unm_niu_gb_mac_config_0_t mac_cfg0;
    if ((port < 0) || (port > UNM_NIU_MAX_GBE_PORTS)) {
        return -1;
    }
    *(unm_crbword_t *)&mac_cfg0 = 0;
    mac_cfg0.soft_reset = 1;
    unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MAC_CONFIG_0(port), &mac_cfg0, 4);
    return 0;
}

/* Disable an XG interface */
native_t unm_niu_disable_xg_port(struct unm_adapter_s *adapter, native_t port) {
    unm_niu_xg_mac_config_0_t mac_cfg;
    if (port != 0) {
        return -1;
    }
    *(unm_crbword_t *)&mac_cfg = 0;
    mac_cfg.soft_reset = 1;
    unm_nic_hw_write_wx(adapter, UNM_NIU_XGE_CONFIG_0, &mac_cfg, 4);
    return 0;
}


/* Set promiscuous mode for a GbE interface */
native_t unm_niu_set_promiscuous_mode(struct unm_adapter_s *adapter,
				      native_t port, unm_niu_prom_mode_t mode) {
    unm_niu_gb_drop_crc_t reg;
    native_t data;
    if ((port < 0) || (port > UNM_NIU_MAX_GBE_PORTS)) {
        return -1;
    }
    if (mode == UNM_NIU_PROMISCOUS_MODE) {
        data = 0;
    } else {
        data = 1;
    }
    /* save previous contents */
    unm_nic_hw_read_wx(adapter, UNM_NIU_GB_DROP_WRONGADDR, &reg, 4);
    switch (port) {
        case 0:
            reg.drop_gb0 = data;
            break;
        case 1:
            reg.drop_gb0 = data;
            break;
        case 2:
            reg.drop_gb0 = data;
            break;
        case 3:
            reg.drop_gb0 = data;
            break;
        default:
            return -1;
    } 
    unm_nic_hw_write_wx(adapter, UNM_NIU_GB_DROP_WRONGADDR, &reg, 4);
    return 0;
}

/*
 * Set the MAC address for an XG port
 * Note that the passed-in value must already be in network byte order.
 */
int
unm_niu_xg_macaddr_set(struct unm_adapter_s *adapter,
		       int phy, unm_ethernet_macaddr_t addr)
{
    unm_crbword_t temp = 0;

    if ((phy < 0) || (phy > 3))
        return -1;

    switch (phy) {
    case 0:
        memcpy(&temp, addr, 2);
        temp <<= 16;
        unm_nic_hw_write_wx(adapter, UNM_NIU_XGE_STATION_ADDR_0_1, &temp, 4);
        temp = 0;
        memcpy(&temp,((__uint8_t *)addr)+2,sizeof(unm_crbword_t));
        unm_nic_hw_write_wx(adapter, UNM_NIU_XGE_STATION_ADDR_0_HI, &temp, 4);
        break;

    case 1:
        memcpy(&temp, addr, 2);
        temp <<= 16;
        unm_nic_hw_write_wx(adapter, UNM_NIU_XG1_STATION_ADDR_0_1, &temp, 4);
        temp = 0;
        memcpy(&temp,((__uint8_t *)addr)+2,sizeof(unm_crbword_t));
        unm_nic_hw_write_wx(adapter, UNM_NIU_XG1_STATION_ADDR_0_HI, &temp, 4);
        break;

    default:
        printk(KERN_ERR "Unknown port %d\n", phy);
        break;
    }

    return 0;
}

/*
 * Return the current station MAC address.
 * Note that the passed-in value must already be in network byte order.
 */
int
unm_niu_xg_macaddr_get(struct unm_adapter_s *adapter,
		       int phy, unm_ethernet_macaddr_t *addr)
{
    unm_crbword_t stationhigh;
    unm_crbword_t stationlow;
    __uint64_t result;

    if (addr == NULL)
        return -1;
    if (phy != 0)
        return -1;

    unm_nic_hw_read_wx(adapter, UNM_NIU_XGE_STATION_ADDR_0_HI, &stationhigh, 4);
    unm_nic_hw_read_wx(adapter, UNM_NIU_XGE_STATION_ADDR_0_1, &stationlow, 4);

    result = ((__uint64_t)stationlow) >> 16;
    result |= (__uint64_t)stationhigh << 16;
    memcpy(*addr,&result,sizeof(unm_ethernet_macaddr_t));

    return 0;
}

native_t
unm_niu_xg_set_promiscuous_mode(struct unm_adapter_s *adapter,
				native_t port, unm_niu_prom_mode_t mode)
{
    long  reg;

    if ((port < 0) || (port > UNM_NIU_MAX_XG_PORTS)) {
        return -1;
    }

    unm_nic_hw_read_wx(adapter, UNM_NIU_XGE_CONFIG_1+(0x10000*port), &reg, 4);
    if (mode == UNM_NIU_PROMISCOUS_MODE) {
        reg = (reg | 0x2000UL);
    } else { /* FIXME  use the correct mode value here*/
        reg = (reg & ~0x2000UL);
    }
    unm_crb_writelit_adapter(adapter, UNM_NIU_XGE_CONFIG_1+(0x10000*port), reg);

    return 0;
}

int
unm_niu_xg_get_tx_flow_ctl(struct unm_adapter_s *adapter,
				int port, int *val)
{
    unm_niu_xg_pause_ctl_t reg;
    if ((port < 0) || (port > UNM_NIU_MAX_XG_PORTS)) {
        return -1;
    }
    unm_nic_hw_read_wx(adapter, UNM_NIU_XG_PAUSE_CTL, &reg, 4);
    if (port == 0) {
        *val = !reg.xg0_mask;
    } else {
        *val = !reg.xg1_mask;
    }
    return 0;
}

int
unm_niu_xg_set_tx_flow_ctl(struct unm_adapter_s *adapter,
				int port, int enable)
{
    unm_niu_xg_pause_ctl_t reg;

    if ((port < 0) || (port > UNM_NIU_MAX_XG_PORTS)) {
        return -1;
    }
    unm_nic_hw_read_wx(adapter, UNM_NIU_XG_PAUSE_CTL, &reg, 4);
    if (port == 0) {
        reg.xg0_mask = !enable;
    } else {
        reg.xg1_mask = !enable;
    }
    unm_nic_hw_write_wx(adapter, UNM_NIU_XG_PAUSE_CTL, &reg, 4);

    return 0;
}

int
unm_niu_gbe_get_tx_flow_ctl(struct unm_adapter_s *adapter,
				int port, int *val)
{
    unm_niu_gb_pause_ctl_t reg;
    if ((port < 0) || (port > UNM_NIU_MAX_GBE_PORTS)) {
        return -1;
    }
    unm_nic_hw_read_wx(adapter, UNM_NIU_GB_PAUSE_CTL, &reg, 4);
    switch (port) {
        case(0):
            *val = !reg.gb0_mask;
            break;
        case(1):
            *val = !reg.gb1_mask;
            break;
        case(2):
            *val = !reg.gb2_mask;
            break;
        case(3):
        default:
            *val = !reg.gb3_mask;
            break;
    }
    return 0;
}
int
unm_niu_gbe_set_tx_flow_ctl(struct unm_adapter_s *adapter,
				int port, int enable)
{
    unm_niu_gb_pause_ctl_t reg;

    if ((port < 0) || (port > UNM_NIU_MAX_GBE_PORTS)) {
        return -1;
    }
    unm_nic_hw_read_wx(adapter, UNM_NIU_GB_PAUSE_CTL, &reg, 4);
    switch (port) {
        case(0):
            reg.gb0_mask = !enable;
            break;
        case(1):
            reg.gb1_mask = !enable;
            break;
        case(2):
            reg.gb2_mask = !enable;
            break;
        case(3):
        default:
            reg.gb3_mask = !enable;
            break;
    }
    unm_nic_hw_write_wx(adapter, UNM_NIU_GB_PAUSE_CTL, &reg, 4);

    return 0;
}

int
unm_niu_gbe_get_rx_flow_ctl(struct unm_adapter_s *adapter,
				int port, int *val)
{
    unm_niu_gb_mac_config_0_t reg;

    if ((port < 0) || (port > UNM_NIU_MAX_GBE_PORTS)) {
        return -1;
    }
    unm_nic_hw_read_wx(adapter, UNM_NIU_GB_MAC_CONFIG_0(port), &reg, 4);
    *val = reg.rx_flowctl;

    return 0;
}
int
unm_niu_gbe_set_rx_flow_ctl(struct unm_adapter_s *adapter,
				int port, int enable)
{
    unm_niu_gb_mac_config_0_t reg;

    if ((port < 0) || (port > UNM_NIU_MAX_GBE_PORTS)) {
        return -1;
    }

    unm_nic_hw_read_wx(adapter, UNM_NIU_GB_MAC_CONFIG_0(port), &reg, 4);
    reg.rx_flowctl = enable;
    unm_nic_hw_write_wx(adapter, UNM_NIU_GB_MAC_CONFIG_0(port), &reg, 4);

    return 0;
}


long unm_niu_xg_enable_phy_interrupts(struct unm_adapter_s *adapter,
				      long port) {
    unm_crb_writelit_adapter(adapter, UNM_NIU_INT_MASK, 0x3f);
    return 0;
}

long unm_niu_xg_disable_phy_interrupts(struct unm_adapter_s *adapter,
				       long port) {
    unm_crb_writelit_adapter(adapter, UNM_NIU_INT_MASK, 0x7f);
    return 0;
}
