/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/* ethtool support for unm nic */

#include <linux/types.h>
#include <asm/uaccess.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <linux/netdevice.h>
#include <linux/ethtool.h>
#include <linux/version.h>

#include "queue.h"
#include "unm_nic_hw.h"
#include "unm_nic.h"
#include "nic_phan_reg.h"
#include "unm_nic_ioctl.h"
#include "nic_cmn.h"
#include "unm_version.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,27) || defined(__VMKERNEL_MODULE__)
extern char * unm_nic_driver_string;
extern void unm_change_ringparam(struct unm_adapter_s *adapter);
extern int rom_fast_read(struct unm_adapter_s *adapter, int addr, int *valp);

#define MIN_RX_DESC_COUNT    4096
#define MIN_TX_DESC_COUNT    4096
#define MAX_RX_DESC_COUNT    MAX_RCV_DESCRIPTORS
#define MAX_TX_DESC_COUNT    MAX_CMD_DESCRIPTORS

// This will be needed for efficiency reasons !!
#define RX_DESC_MULTIPLE    8
#define TX_DESC_MULTIPLE    8

#define UNM_ROUNDUP(i, size)    ((i) = (((i) + (size) - 1) & ~((size) - 1)))

struct unm_nic_stats {
        char stat_string[ETH_GSTRING_LEN];
        int sizeof_stat;
        int stat_offset;
};

#define UNM_NIC_STAT(m) sizeof(((struct unm_adapter_s *)0)->m), \
                    offsetof(struct unm_adapter_s, m)

static const struct unm_nic_stats unm_nic_gstrings_stats[] = {
        { "rcvd bad skb",     UNM_NIC_STAT(stats.rcvdbadskb) },
        { "xmit called",      UNM_NIC_STAT(stats.xmitcalled) },
        { "xmited frames",    UNM_NIC_STAT(stats.xmitedframes) },
        { "xmit finished",    UNM_NIC_STAT(stats.xmitfinished) },
        { "bad skb len",      UNM_NIC_STAT(stats.badskblen) },
        { "no cmd desc",      UNM_NIC_STAT(stats.nocmddescriptor) },
        { "polled",           UNM_NIC_STAT(stats.polled) },
        { "uphappy",          UNM_NIC_STAT(stats.uphappy) },
        { "updropped",        UNM_NIC_STAT(stats.updropped) },
        { "uplcong",          UNM_NIC_STAT(stats.uplcong) },
        { "uphcong",          UNM_NIC_STAT(stats.uphcong) },
        { "upmcong",          UNM_NIC_STAT(stats.upmcong) },
        { "updunno",          UNM_NIC_STAT(stats.updunno) },
        { "skb freed",        UNM_NIC_STAT(stats.skbfreed) },
        { "tx dropped",       UNM_NIC_STAT(stats.txdropped) },
        { "tx null skb",      UNM_NIC_STAT(stats.txnullskb) },
        { "csummed",          UNM_NIC_STAT(stats.csummed) },
        { "no rcv",           UNM_NIC_STAT(stats.no_rcv) },
        { "rx bytes",         UNM_NIC_STAT(stats.rxbytes) },
        { "tx bytes",         UNM_NIC_STAT(stats.txbytes) },
};

#define UNM_NIC_STATS_LEN        \
        sizeof(unm_nic_gstrings_stats) / sizeof(struct unm_nic_stats)

static const char unm_nic_gstrings_test[][ETH_GSTRING_LEN] = {
        "Register test  (on/offline)",
#if 0
        "Eeprom test    (offline)",
        "Interrupt test (offline)",
        "Loopback test  (offline)",
#endif
        "Link test   (on/offline)"
};
#define UNM_NIC_TEST_LEN sizeof(unm_nic_gstrings_test) / ETH_GSTRING_LEN

#define UNM_NIC_REGS_COUNT 42
#define UNM_NIC_REGS_LEN (UNM_NIC_REGS_COUNT * sizeof(unm_crbword_t))

#define TRUE    1
#define FALSE   0

static int
unm_nic_get_eeprom_len(struct net_device *netdev)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);
        int n;
        if ((rom_fast_read(adapter,0, &n) == 0) &&
            (n & 0x80000000)) { // verify the rom address
                n &= ~0x80000000;
                if(n < 1024) return (n);
        }
        return 0;
}

static void
unm_nic_get_drvinfo(struct net_device *netdev, struct ethtool_drvinfo *drvinfo)
{
        struct unm_adapter_s   *adapter = netdev_priv(netdev);
        unsigned long          flags;
        uint32_t               fw_major = 0;
        uint32_t               fw_minor = 0;
        uint32_t               fw_build = 0;

        strncpy(drvinfo->driver, DRIVER_NAME, 32);
        strncpy(drvinfo->version, UNM_NIC_LINUX_VERSIONID, 32);

        UNM_WRITE_LOCK_IRQS(&adapter->adapter_lock, flags);
		fw_major = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
				UNM_FW_VERSION_MAJOR));
		fw_minor = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
				UNM_FW_VERSION_MINOR));
		fw_build = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
				UNM_FW_VERSION_SUB));
        UNM_WRITE_UNLOCK_IRQR(&adapter->adapter_lock, flags);
        sprintf(drvinfo->fw_version, "%d.%d.%d", fw_major, fw_minor,
             fw_build);

        strncpy(drvinfo->bus_info, pci_name(adapter->pdev), 32);
        drvinfo->n_stats = UNM_NIC_STATS_LEN;
        drvinfo->testinfo_len = UNM_NIC_TEST_LEN;
        drvinfo->regdump_len = UNM_NIC_REGS_LEN;
        drvinfo->eedump_len = unm_nic_get_eeprom_len(netdev);
}

static int
unm_nic_get_settings(struct net_device *netdev, struct ethtool_cmd *ecmd)
{
        struct unm_adapter_s   *adapter = netdev_priv(netdev);
        unm_board_info_t       *boardinfo;

        boardinfo = &adapter->ahw.boardcfg;

        // read which mode
        if(adapter->ahw.board_type == UNM_NIC_GBE) {
                ecmd->supported  =  (SUPPORTED_10baseT_Half|
                                    SUPPORTED_10baseT_Full|
                                    SUPPORTED_100baseT_Half|
                                    SUPPORTED_100baseT_Full|
                                    SUPPORTED_1000baseT_Half|
                                     SUPPORTED_1000baseT_Full);

                ecmd->advertising = (ADVERTISED_100baseT_Half|
                                        ADVERTISED_100baseT_Full|
                                        ADVERTISED_1000baseT_Half|
                                     ADVERTISED_1000baseT_Full);

//        if(netif_carrier_ok(netdev)) { // cant use this method to determine
//        link status as driver has not taken care of carrier state (ON/OFF).
//        Instead I'll use cached value in unm_port to find link status !!
                if(adapter->state) {
                        ecmd->speed = adapter->link_speed;
                        ecmd->duplex = adapter->link_duplex;
                } else
                        return -EIO; // link absent

        } else if (adapter->ahw.board_type == UNM_NIC_XGBE) {
                ecmd->supported   =  SUPPORTED_10000baseT_Full;
                ecmd->advertising = ADVERTISED_10000baseT_Full;
                ecmd->speed       = SPEED_10000;
                ecmd->duplex      = DUPLEX_FULL;
        } else {
            printk (KERN_ERR "%s: ERROR: Unsupported board model %d\n",
                     unm_nic_driver_name, (unm_brdtype_t)boardinfo->board_type);
            return -EIO;
        }

                ecmd->phy_address = adapter->portnum;
                ecmd->transceiver = XCVR_EXTERNAL;

        switch ((unm_brdtype_t)boardinfo->board_type) {
            case UNM_BRDTYPE_P2_SB35_4G:
            case UNM_BRDTYPE_P2_SB31_2G:
                ecmd->supported   |= SUPPORTED_Autoneg;
                ecmd->advertising |= ADVERTISED_Autoneg;
            case UNM_BRDTYPE_P2_SB31_10G_CX4:
                ecmd->supported   |= SUPPORTED_TP;
                ecmd->advertising |= ADVERTISED_TP;
                ecmd->port         = PORT_TP;
                ecmd->autoneg      = (boardinfo->board_type ==
                                      UNM_BRDTYPE_P2_SB31_10G_CX4) ?
                                      (AUTONEG_DISABLE) : (adapter->link_autoneg);
                break;
            case UNM_BRDTYPE_P2_SB31_10G_HMEZ:
            case UNM_BRDTYPE_P2_SB31_10G_IMEZ:
                ecmd->supported   |= SUPPORTED_MII;
                ecmd->advertising |= ADVERTISED_MII;
                ecmd->port         = PORT_FIBRE;
                ecmd->autoneg      = AUTONEG_DISABLE;
                break;
            case UNM_BRDTYPE_P2_SB31_10G:
                ecmd->supported   |= SUPPORTED_FIBRE;
                ecmd->advertising |= ADVERTISED_FIBRE;
                ecmd->port         = PORT_FIBRE;
                ecmd->autoneg      = AUTONEG_DISABLE;
                break;
            default: printk (KERN_ERR "%s: ERROR: Unsupported board model %d\n",
                         unm_nic_driver_name, (unm_brdtype_t)boardinfo->board_type);
        return -EIO;
                break;
        }

        return 0;
}

static int
unm_nic_set_settings(struct net_device *netdev, struct ethtool_cmd *ecmd)
{
        struct unm_adapter_s   *adapter = netdev_priv(netdev);
        unm_niu_phy_status_t    status;

        // read which mode
        if(adapter->ahw.board_type == UNM_NIC_GBE) {
                // autonegotiation
                if (unm_nic_phy_write(adapter,physical_port[adapter->portnum],
                                UNM_NIU_GB_MII_MGMT_ADDR_AUTONEG,
                                (unm_crbword_t)ecmd->autoneg)!= 0)
                        return -EIO;
                else
                        adapter->link_autoneg = ecmd->autoneg;

                if (unm_nic_phy_read(adapter, physical_port[adapter->portnum],
                                UNM_NIU_GB_MII_MGMT_ADDR_PHY_STATUS,
                                (unm_crbword_t *)&status) != 0)
                        return -EIO;
                // speed
                switch(ecmd->speed) {
                case SPEED_10:
                        status.speed = 0;
                        break;
                case SPEED_100:
                        status.speed = 1;
                        break;
                case SPEED_1000:
                        status.speed = 2;
                        break;
                }
                // set duplex mode
                if(ecmd->duplex == DUPLEX_HALF)
                        status.duplex = 0;
                if(ecmd->duplex == DUPLEX_FULL)
                        status.duplex = 1;

                if (unm_nic_phy_write(adapter,physical_port[adapter->portnum],
                        UNM_NIU_GB_MII_MGMT_ADDR_PHY_STATUS,
                        *((int*)&status)) != 0)
                        return -EIO;
                else {
                        adapter->link_speed = ecmd->speed;
                        adapter->link_duplex = ecmd->duplex;
                }
        } else if (adapter->ahw.board_type == UNM_NIC_XGBE) {
            // no changes supported for XGBE
                return -EIO;
        }
        if(netif_running(netdev)) {
                netdev->stop(netdev);
                netdev->open(netdev);
        }
        return 0;
}

static int
unm_nic_get_regs_len(struct net_device *netdev)
{
        return UNM_NIC_REGS_LEN;
}

static void
unm_nic_get_regs(struct net_device *netdev, struct ethtool_regs *regs, void *p)
{
        struct unm_adapter_s   *adapter = netdev_priv(netdev);
        unm_crbword_t           mode, *regs_buff = p;
        unsigned long           flags;
        void                   *addr;

        memset(p, 0, UNM_NIC_REGS_LEN);
        regs->version = (1 << 24) | (adapter->ahw.revision_id << 16) |
                            adapter->ahw.device_id;
	UNM_WRITE_LOCK_IRQS(&adapter->adapter_lock, flags);
        // which mode
        UNM_NIC_LOCKED_READ_REG(UNM_NIU_MODE, &regs_buff[0]);
        mode = regs_buff[0];

        // Common registers to all the modes
        UNM_NIC_LOCKED_READ_REG(UNM_NIU_STRAP_VALUE_SAVE_HIGHER, &regs_buff[2]);
        switch(mode) {
        case 4: {//XGB Mode
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_SINGLE_TERM, &regs_buff[3]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_DRIVE_HI, &regs_buff[4]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_DRIVE_LO, &regs_buff[5]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_DTX, &regs_buff[6]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_DEQ, &regs_buff[7]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_WORD_ALIGN, &regs_buff[8]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_RESET, &regs_buff[9]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_POWER_DOWN, &regs_buff[10]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_RESET_PLL, &regs_buff[11]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_SERDES_LOOPBACK,
                                        &regs_buff[12]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_DO_BYTE_ALIGN,
                                        &regs_buff[13]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_TX_ENABLE, &regs_buff[14]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_RX_ENABLE, &regs_buff[15]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_STATUS, &regs_buff[16]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XG_PAUSE_THRESHOLD,
                                        &regs_buff[17]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_CONFIG_0, &regs_buff[18]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_CONFIG_1, &regs_buff[19]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_IPG, &regs_buff[20]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_STATION_ADDR_0_HI,
                                        &regs_buff[21]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_STATION_ADDR_0_1,
                                        &regs_buff[22]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_STATION_ADDR_1_LO,
                                        &regs_buff[23]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_STATUS, &regs_buff[24]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_MAX_FRAME_SIZE,
                                        &regs_buff[25]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_PAUSE_FRAME_VALUE,
                                        &regs_buff[26]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_TX_BYTE_CNT,
                                        &regs_buff[27]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_TX_FRAME_CNT,
                                        &regs_buff[28]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_RX_BYTE_CNT,
                                        &regs_buff[29]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_RX_FRAME_CNT,
                                        &regs_buff[30]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_AGGR_ERROR_CNT,
                                        &regs_buff[31]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_MULTICAST_FRAME_CNT,
                                        &regs_buff[32]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_UNICAST_FRAME_CNT,
                                        &regs_buff[33]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_CRC_ERROR_CNT,
                                        &regs_buff[34]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_OVERSIZE_FRAME_ERR,
                                        &regs_buff[35]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_UNDERSIZE_FRAME_ERR,
                                        &regs_buff[36]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_LOCAL_ERROR_CNT,
                                        &regs_buff[37]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_REMOTE_ERROR_CNT,
                                        &regs_buff[38]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_CONTROL_CHAR_CNT,
                                        &regs_buff[39]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_PAUSE_FRAME_CNT,
                                        &regs_buff[40]);
                break;
                }

        case 2: {  // GB Mode
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_SERDES_RESET, &regs_buff[3]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB0_MII_MODE, &regs_buff[4]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB1_MII_MODE, &regs_buff[5]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB2_MII_MODE, &regs_buff[6]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB3_MII_MODE, &regs_buff[7]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB0_GMII_MODE, &regs_buff[8]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB1_GMII_MODE, &regs_buff[9]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB2_GMII_MODE, &regs_buff[10]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB3_GMII_MODE, &regs_buff[11]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_REMOTE_LOOPBACK,
                                        &regs_buff[12]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB0_HALF_DUPLEX,
                                        &regs_buff[13]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB1_HALF_DUPLEX,
                                        &regs_buff[14]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_RESET_SYS_FIFOS,
                                        &regs_buff[15]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_CRC_DROP,
                                        &regs_buff[16]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_DROP_WRONGADDR,
                                        &regs_buff[17]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_TEST_MUX_CTL, &regs_buff[18]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_MAC_CONFIG_0(physical_port[adapter->portnum]),
                                        &regs_buff[19]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_MAC_CONFIG_1(physical_port[adapter->portnum]),
                                        &regs_buff[20]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_HALF_DUPLEX_CTRL(physical_port[adapter->portnum]),
                                        &regs_buff[21]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_MAX_FRAME_SIZE(physical_port[adapter->portnum]),
                                        &regs_buff[22]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_TEST_REG(physical_port[adapter->portnum]),
                                        &regs_buff[23]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_MII_MGMT_CONFIG(physical_port[adapter->portnum]),
                                        &regs_buff[24]);
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_MII_MGMT_COMMAND(physical_port[adapter->portnum]),
                                        &regs_buff[25]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_MII_MGMT_ADDR(physical_port[adapter->portnum]),
                                        &regs_buff[26]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_MII_MGMT_CTRL(physical_port[adapter->portnum]),
                                        &regs_buff[27]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_MII_MGMT_STATUS(physical_port[adapter->portnum]),
                                        &regs_buff[28]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_MII_MGMT_INDICATE(physical_port[adapter->portnum]),
                                        &regs_buff[29]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_INTERFACE_CTRL(physical_port[adapter->portnum]),
                                        &regs_buff[30]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_INTERFACE_STATUS(physical_port[adapter->portnum]),
                                        &regs_buff[31]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_STATION_ADDR_0(physical_port[adapter->portnum]),
                                        &regs_buff[32]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_GB_STATION_ADDR_1(physical_port[adapter->portnum]),
                                        &regs_buff[33]);
                break;
                }

        case 1: { // FC Mode
                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_RX_STATUS(physical_port[adapter->portnum]),
                            &regs_buff[3]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_RX_COMMA_DETECT(physical_port[adapter->portnum]),
                            &regs_buff[4]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_LASER_UNSAFE(physical_port[adapter->portnum]),
                            &regs_buff[5]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_TX_CONTROL(physical_port[adapter->portnum]),
                            &regs_buff[6]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_ON_OFFLINE_CTL(physical_port[adapter->portnum]),
                            &regs_buff[7]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_PORT_ACTIVE_STAT(physical_port[adapter->portnum]),
                            &regs_buff[8]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_PORT_INACTIVE_STAT(physical_port[adapter->portnum]),
                            &regs_buff[9]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_LINK_FAILURE_CNT(physical_port[adapter->portnum]),
                            &regs_buff[10]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_LOSS_SYNC_CNT(physical_port[adapter->portnum]),
                            &regs_buff[11]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_LOSS_SIGNAL_CNT(physical_port[adapter->portnum]),
                            &regs_buff[12]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_PRIM_SEQ_ERR_CNT(physical_port[adapter->portnum]),
                            &regs_buff[13]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_INVLD_TX_WORD_CNT(physical_port[adapter->portnum]),
                            &regs_buff[14]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_INVLD_CRC_CNT(physical_port[adapter->portnum]),
                            &regs_buff[15]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_RX_CELL_CNT(physical_port[adapter->portnum]),
                            &regs_buff[16]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_TX_CELL_CNT(physical_port[adapter->portnum]),
                            &regs_buff[17]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_B2B_CREDIT(physical_port[adapter->portnum]),
                            &regs_buff[18]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_LOGIN_DONE(physical_port[adapter->portnum]),
                            &regs_buff[19]);

                UNM_NIC_LOCKED_READ_REG(UNM_NIU_FC_OPERATING_SPEED(physical_port[adapter->portnum]),
                            &regs_buff[20]);
                break;
                }
        }
	UNM_WRITE_UNLOCK_IRQR(&adapter->adapter_lock, flags);
}

static uint32_t
unm_nic_get_msglevel(struct net_device *netdev)
{
        /* Presently unm_nic_debug=0 and is not being used to display
         * debug msgs. But comment says that it will be used in future
         * releases to display debug msgs(which is exactly the way it
         * should be) Keeping this thing in mind !!
         */
        return unm_nic_debug;
}

static void
unm_nic_set_msglevel(struct net_device *netdev, uint32_t data)
{
        unm_nic_debug = data;
}

/* Restart Link Process */
static int
unm_nic_nway_reset(struct net_device *netdev)
{
        if(netif_running(netdev)) {
                netdev->stop(netdev); // verify
                netdev->open(netdev);
        }
        return 0;
}

static int
unm_nic_test_link(struct net_device *netdev)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);
        int                  link = 1;

        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE: {
                unm_niu_phy_status_t status;
                if (unm_nic_phy_read(adapter, physical_port[adapter->portnum],
                        UNM_NIU_GB_MII_MGMT_ADDR_PHY_STATUS,
                      (unm_crbword_t *)&status) == 0)
                    link = (status.link == 1 ? 0 : 1);
                break;
        }

        case UNM_NIC_XGBE:
                link =
                   ((xge_link_status(adapter)==XG_LINK_UP) ? (0) : (1));
                break;
        }
        return(link);
}

static int
unm_nic_get_eeprom(struct net_device *netdev,
                struct ethtool_eeprom *eeprom, uint8_t *bytes)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);
        int i;

        if(eeprom->len == 0) return -EINVAL;

        eeprom->magic = adapter->ahw.vendor_id |
                        (adapter->ahw.device_id << 16);
        for(i=0; i<eeprom->len; i++)
                if(rom_fast_read(adapter,(8*i)+8, (int *)eeprom->data) == -1)
                        return -1;
        return 0;
}

static void
unm_nic_get_ringparam(struct net_device *netdev,
                        struct ethtool_ringparam *ring)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);
        int i, j;

        ring->rx_pending = 0;
        for (i = 0; i < MAX_RCV_CTX; ++i) {
                for (j = 0; j < NUM_RCV_DESC_RINGS; j++) {
                        ring->rx_pending +=
                                adapter->recv_ctx[i].rcv_desc[j].rcv_pending;
                }
        }

        ring->tx_pending = adapter->pendingCmdCount;
        ring->rx_max_pending = adapter->MaxRxDescCount;
        ring->tx_max_pending = adapter->MaxTxDescCount;
        ring->rx_mini_max_pending = 0;
        ring->rx_mini_pending = 0;
        ring->rx_jumbo_max_pending = 0;
        ring->rx_jumbo_pending = 0;
}

// Note: This change will be reflected in all the four ports as there is
// only one common adapter.
// NO CHANGE of ringparams allowed !
static int
unm_nic_set_ringparam(struct net_device *netdev,
                        struct ethtool_ringparam *ring)
{
        return(-EIO);
}

static void
unm_nic_get_pauseparam(struct net_device *netdev,
                    struct ethtool_pauseparam *pause)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);
        int temp;
        if(adapter->ahw.board_type == UNM_NIC_GBE) {
            if (unm_niu_gbe_get_rx_flow_ctl(adapter, physical_port[adapter->portnum], &temp)==0)
                pause->rx_pause = temp;
            if (unm_niu_gbe_get_tx_flow_ctl(adapter, physical_port[adapter->portnum], &temp)==0)
                pause->tx_pause = temp;

        } else if(adapter->ahw.board_type == UNM_NIC_XGBE) {
            if (unm_niu_xg_get_tx_flow_ctl(adapter, physical_port[adapter->portnum], &temp)==0) {
                pause->tx_pause = temp;
            }
            pause->rx_pause = 1;   // always on for XG
        } else {
            printk(KERN_ERR "%s: Unknown board type: %x\n",unm_nic_driver_name,
                   adapter->ahw.board_type);
        }

}

static int
unm_nic_set_pauseparam(struct net_device *netdev,
                            struct ethtool_pauseparam *pause)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);
	int ret = 0;

    if(adapter->ahw.board_type == UNM_NIC_GBE) {
        if (unm_niu_gbe_set_rx_flow_ctl(adapter, physical_port[adapter->portnum],
                                             pause->rx_pause)){
            ret= -EIO;
        }
        if (unm_niu_gbe_set_tx_flow_ctl(adapter, physical_port[adapter->portnum],
                                             pause->tx_pause)){
            ret= -EIO;
        }
    } else if(adapter->ahw.board_type == UNM_NIC_XGBE) {
        if (unm_niu_xg_set_tx_flow_ctl(adapter, physical_port[adapter->portnum],
                                             pause->tx_pause)){
            ret =  -EIO;
        }
    } else {
        ret= -EIO;
    }
    return ret;
}

static uint32_t
unm_nic_get_rx_csum(struct net_device *netdev)
{
        struct unm_adapter_s *adapter = (struct unm_adapter_s *)(netdev->priv);
        return adapter->rx_csum;
}

static int
unm_nic_set_rx_csum(struct net_device *netdev, uint32_t data)
{
	struct unm_adapter_s *adapter = (struct unm_adapter_s *)(netdev->priv);
        adapter->rx_csum = data;
        return 0;
}

static int
unm_nic_reg_test(struct net_device *netdev)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);
        uint32_t data_read, data_written;

        // Read test
        unm_nic_read_w0(adapter, UNM_PCIX_PH_REG(0), &data_read);
        if ((data_read & 0xffff) != PHAN_VENDOR_ID) {
            return 1;
        }

        // write test
        data_written = (uint32_t)0xa5a5a5a5;

        unm_nic_reg_write(adapter, CRB_SCRATCHPAD_TEST, data_written);
        data_read = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
				CRB_SCRATCHPAD_TEST));
        if (data_written != data_read) {
            return 1;
        }

        return 0;
}

static int
unm_nic_diag_test_count(struct net_device *netdev)
{
        return UNM_NIC_TEST_LEN;
}

static void
unm_nic_diag_test(struct net_device *netdev,
                struct ethtool_test *eth_test, uint64_t *data)
{
        if(eth_test->flags == ETH_TEST_FL_OFFLINE) { // offline tests
                // link test
                if((data[1] = (uint64_t)unm_nic_test_link(netdev)))
                        eth_test->flags |= ETH_TEST_FL_FAILED;

                // register tests
                if((data[0] = unm_nic_reg_test(netdev)))
                        eth_test->flags |= ETH_TEST_FL_FAILED;

        } else { // online tests
                // register tests
                if((data[0] = unm_nic_reg_test(netdev)))
                        eth_test->flags |= ETH_TEST_FL_FAILED;
                // link test
                if((data[1] = (uint64_t)unm_nic_test_link(netdev)))
                        eth_test->flags |= ETH_TEST_FL_FAILED;

        }
}

static void
unm_nic_get_strings(struct net_device *netdev, uint32_t stringset,
                            uint8_t *data)
{
        int i;

        switch(stringset) {
        case ETH_SS_TEST:
                memcpy(data, *unm_nic_gstrings_test,
                UNM_NIC_TEST_LEN*ETH_GSTRING_LEN);
                break;
        case ETH_SS_STATS:
                for (i=0; i < UNM_NIC_STATS_LEN; i++) {
                        memcpy(data + i * ETH_GSTRING_LEN,
                        unm_nic_gstrings_stats[i].stat_string,
                        ETH_GSTRING_LEN);
                }
                break;
        }
}

static int
unm_nic_get_stats_count(struct net_device *netdev)
{
        return UNM_NIC_STATS_LEN;
}

/*
 * NOTE: I have displayed only port's stats
 * TBD: unm_nic_stats(struct unm_port * port) doesn't update stats
 * as of now !! So this function may produce unexpected results !!
 */
#if 0
static void
unm_nic_get_ethtool_stats(struct net_device *netdev,
            struct ethtool_stats *stats, uint64_t *data)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);
        int i;

        for(i = 0; i < UNM_NIC_STATS_LEN; i++) {
                char *p = (char *)port+unm_nic_gstrings_stats[i].stat_offset;
                data[i] = (unm_nic_gstrings_stats[i].sizeof_stat ==
                        sizeof(uint64_t)) ? *(uint64_t *)p : *(uint32_t *)p;
        }

}
#endif

static struct ethtool_ops unm_nic_ethtool_ops = {
        .get_settings        = unm_nic_get_settings,
        .set_settings        = unm_nic_set_settings,
        .get_drvinfo         = unm_nic_get_drvinfo,
        .get_regs_len        = unm_nic_get_regs_len,
        .get_regs            = unm_nic_get_regs,
        .get_msglevel        = unm_nic_get_msglevel,
        .set_msglevel        = unm_nic_set_msglevel,
        .nway_reset          = unm_nic_nway_reset,
        .get_link            = ethtool_op_get_link,

#ifndef __VMKERNEL_MODULE__
        .get_eeprom_len      = unm_nic_get_eeprom_len,
#endif
        .get_eeprom          = unm_nic_get_eeprom,
        .get_ringparam       = unm_nic_get_ringparam,
        .set_ringparam       = unm_nic_set_ringparam,
        .get_pauseparam      = unm_nic_get_pauseparam,
        .set_pauseparam      = unm_nic_set_pauseparam,
        .get_rx_csum         = unm_nic_get_rx_csum,
        .set_rx_csum         = unm_nic_set_rx_csum,
        .get_tx_csum         = ethtool_op_get_tx_csum,
#ifndef __VMKERNEL_MODULE__
        .set_tx_csum         = ethtool_op_set_tx_csum,
#endif
        .get_sg              = ethtool_op_get_sg,
        .set_sg              = ethtool_op_set_sg,
#ifndef __VMKERNEL_MODULE__
#ifdef NETIF_F_TSO
        .get_tso             = ethtool_op_get_tso,
        .set_tso             = ethtool_op_set_tso,
#endif
#endif
        .self_test_count     = unm_nic_diag_test_count,
        .self_test           = unm_nic_diag_test,
        .get_strings         = unm_nic_get_strings,
        .get_stats_count     = unm_nic_get_stats_count,
#if 0
        .get_ethtool_stats   = unm_nic_get_ethtool_stats,
#endif
};

void set_ethtool_ops(struct net_device *netdev)
{
        netdev->ethtool_ops = &unm_nic_ethtool_ops;
}
#endif /* LINUX_VERSION */
